#include <math.h>
#include <stdio.h>

#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include "Prevaricator.h"



#define ER_MEM 				0xe000
#define ER_CANNOTOPEN 		0xe100
#define ER_NOTWAVEFILE 		0xe101
#define ER_CANNOTREAD 		0xe102
#define ER_CORRUPTWAVEFILE	0xe103
#define ER_CANNOTWRITE		0xe104
#define ASSERT(x)

#define ID_TIMER    1

//default recording rate
#define DEFAULT_SAMPLE_RATE 11025

//# ticks allowed in PLAYBACK scrollbar
#define PLAYBACK_RESOLUTION 100


#define UNTITLED TEXT ("(untitled)")


// Variables resulting from analysis
BOOL SampleValid;
double Average_Volume;
double Average_Frequency;
double Lowest_Frequency;
double Highest_Frequency;
double Frequency_Range;
double Frequency_Threshold;

#define GRAPHPOINTCOUNT 60
POINT GraphPoints[GRAPHPOINTCOUNT] ;
int GRX;
int GRY;


TCHAR szAppName[]     = TEXT("Prevaricator") ;
TCHAR szBusyError[]   = TEXT("File Operations Disallowed when in play/record!");
TCHAR szNoData[]      = TEXT("No Data to Save!");
TCHAR szFormat[]      = TEXT("Supports only 16 Bit Mono Wave Audio Formats!");
TCHAR szReformat[]    = TEXT("Convert 8 Bit Audio to 16 Bit?");
TCHAR szOutOfMemory[] = TEXT("Out of Memory!\nAllocated:%d");
TCHAR szOpenErrorDev[] = TEXT ("Error opening waveform audio!\nDevice Capability:%d");
TCHAR szOpenError[] = TEXT ("Error opening waveform audio!");


HWND  hwndModeless ;
HWND  hwndMessage ;
HWND  hwndPlayback ;
HWND  hwndGraph;
HWND  hwndBuddy;
HWND  hwndSpinCtl;

BOOL  bNeedSave;
BOOL  bRecording, bPlaying;


// WAVEFILE header
static WAVEFORMATEX waveform ;

// in-memory wavefile audio data location
static PBYTE pSaveBuffer ;

// length and current location within audio data
static int dwDataLength, dwCurrentPos ;


static OPENFILENAME ofn ;

static PSTR   CmdLine;
static int    CmdShow;
static TCHAR  szFileName[MAX_PATH], szTitleName[MAX_PATH] ;


LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MainDlg (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) ;


int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    PSTR szCmdLine, int iCmdShow)
{
     CmdLine = szCmdLine;
     CmdShow = iCmdShow;
     if (-1 == DialogBox (hInstance, szAppName, NULL, MainDlg))
     {
          MessageBox (NULL, TEXT ("This program requires Win95 or better"),
                      szAppName, MB_ICONERROR) ;
     }
     return 0 ;
}


short AskAboutSave (HWND hwnd, TCHAR * szTitleName)
{
     TCHAR szBuffer[64 + MAX_PATH] ;
     int   iReturn ;
     
     wsprintf (szBuffer, TEXT ("Save current changes in %s?"),
               szTitleName[0] ? szTitleName : UNTITLED) ;
     
     iReturn = MessageBox (hwnd, szBuffer, szAppName,
                           MB_YESNOCANCEL | MB_ICONQUESTION) ;
     
     if (iReturn == IDYES)
          if (!SendMessage (hwnd, WM_COMMAND, IDM_FILE_SAVE, 0))
               iReturn = IDCANCEL ;
          
     return iReturn ;
}

void DoCaption (HWND hwnd, TCHAR * szTitleName)
{
     TCHAR szCaption[64 + MAX_PATH] ;
     
     wsprintf (szCaption, TEXT ("%s - %s"), szAppName,
               szTitleName[0] ? szTitleName : UNTITLED) ;
     
     SetWindowText (hwnd, szCaption) ;
}

void OkMessage (HWND hwnd, TCHAR * szMessage, TCHAR * szTitleName)
{
     TCHAR szBuffer[64 + MAX_PATH] ;
     
     wsprintf (szBuffer, szMessage, szTitleName[0] ? szTitleName : UNTITLED) ;
     
     MessageBox (hwnd, szBuffer, szAppName, MB_OK | MB_ICONEXCLAMATION) ;
}

void PopFileInitialize (HWND hwnd)
{
     // ONLY MONO .WAV files are supported,
     // but we will allow *.* in case something is named funny
     static TCHAR szFilter[] = TEXT ("Sound (*.wav)\0*.wav\0")  \
                               TEXT ("All Files (*.*)\0*.*\0\0") ;
     
     ofn.lStructSize       = sizeof (OPENFILENAME) ;
     ofn.hwndOwner         = hwnd ;
     ofn.hInstance         = NULL ;
     ofn.lpstrFilter       = szFilter ;
     ofn.lpstrCustomFilter = NULL ;
     ofn.nMaxCustFilter    = 0 ;
     ofn.nFilterIndex      = 0 ;
     ofn.lpstrFile         = NULL ;          // Set in Open and Close functions
     ofn.nMaxFile          = MAX_PATH ;
     ofn.lpstrFileTitle    = NULL ;          // Set in Open and Close functions
     ofn.nMaxFileTitle     = MAX_PATH ;
     ofn.lpstrInitialDir   = NULL ;
     ofn.lpstrTitle        = NULL ;
     ofn.Flags             = 0 ;             // Set in Open and Close functions
     ofn.nFileOffset       = 0 ;
     ofn.nFileExtension    = 0 ;
     ofn.lpstrDefExt       = TEXT ("txt") ;
     ofn.lCustData         = 0L ;
     ofn.lpfnHook          = NULL ;
     ofn.lpTemplateName    = NULL ;
}

int WaveOpenFile(
	TCHAR*pszFileName,                              // (IN)
	HMMIO *phmmioIn,                                // (OUT)
	WAVEFORMATEX **ppwfxInfo,                       // (OUT)
	MMCKINFO *pckInRIFF                             // (OUT)
			)
{
	HMMIO           hmmioIn;
	MMCKINFO        ckIn;           // chunk info. for general use.
	PCMWAVEFORMAT   pcmWaveFormat;  // Temp PCM structure to load in.       
	WORD            cbExtraAlloc;   // Extra bytes for waveformatex 
	int             nError;         // Return value.
	// Initialization...
	*ppwfxInfo = NULL;
	nError = 0;
	hmmioIn = NULL;
		if ((hmmioIn = mmioOpen(pszFileName, NULL, MMIO_ALLOCBUF | MMIO_READ)) == NULL)
		{
		nError = ER_CANNOTOPEN;
		goto ERROR_READING_WAVE;
		}
	if ((nError = (int)mmioDescend(hmmioIn, pckInRIFF, NULL, 0)) != 0)
		{
		goto ERROR_READING_WAVE;
		}
	if ((pckInRIFF->ckid != FOURCC_RIFF) || (pckInRIFF->fccType != mmioFOURCC('W', 'A', 'V', 'E')))
		{
		nError = ER_NOTWAVEFILE;
		goto ERROR_READING_WAVE;
		}			
	/* Search the input file for for the 'fmt ' chunk.     */
    ckIn.ckid = mmioFOURCC('f', 'm', 't', ' ');
    if ((nError = (int)mmioDescend(hmmioIn, &ckIn, pckInRIFF, MMIO_FINDCHUNK)) != 0)
		{
		goto ERROR_READING_WAVE;                
		}			
	/* Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>;
    * if there are extra parameters at the end, we'll ignore them */
    if (ckIn.cksize < (long) sizeof(PCMWAVEFORMAT))
		{
		nError = ER_NOTWAVEFILE;
		goto ERROR_READING_WAVE;
		}															
	/* Read the 'fmt ' chunk into <pcmWaveFormat>.*/     
    if (mmioRead(hmmioIn, (HPSTR) &pcmWaveFormat, (long) sizeof(pcmWaveFormat)) != (long) sizeof(pcmWaveFormat))
		{
		nError = ER_CANNOTREAD;
		goto ERROR_READING_WAVE;
		}
	// Ok, allocate the waveformatex, but if its not pcm
	// format, read the next word, and thats how many extra
	// bytes to allocate.
	if (pcmWaveFormat.wf.wFormatTag == WAVE_FORMAT_PCM)
		cbExtraAlloc = 0;                               					
	else
		{
		// Read in length of extra bytes.
		if (mmioRead(hmmioIn, (LPTSTR) &cbExtraAlloc,
			(long) sizeof(cbExtraAlloc)) != (long) sizeof(cbExtraAlloc))
			{
			nError = ER_CANNOTREAD;
			goto ERROR_READING_WAVE;
			}
		}
	// Ok, now allocate that waveformatex structure.
	if ((*ppwfxInfo = GlobalAlloc(GMEM_FIXED, sizeof(WAVEFORMATEX)+cbExtraAlloc)) == NULL)
		{
		nError = ER_MEM;
		goto ERROR_READING_WAVE;
		}
	// Copy the bytes from the pcm structure to the waveformatex structure
	memcpy(*ppwfxInfo, &pcmWaveFormat, sizeof(pcmWaveFormat));
	(*ppwfxInfo)->cbSize = cbExtraAlloc;
	// Now, read those extra bytes into the structure, if cbExtraAlloc != 0.
	if (cbExtraAlloc != 0)
		{
		if (mmioRead(hmmioIn, (LPTSTR) (((BYTE*)&((*ppwfxInfo)->cbSize))+sizeof(cbExtraAlloc)),
			(long) (cbExtraAlloc)) != (long) (cbExtraAlloc))
			{
			nError = ER_NOTWAVEFILE;
			goto ERROR_READING_WAVE;
			}
		}
	/* Ascend the input file out of the 'fmt ' chunk. */                                                            
	if ((nError = mmioAscend(hmmioIn, &ckIn, 0)) != 0)
		{
		goto ERROR_READING_WAVE;
		}
	goto TEMPCLEANUP;               

ERROR_READING_WAVE:
	if (*ppwfxInfo != NULL)
		{
		GlobalFree(*ppwfxInfo);
		*ppwfxInfo = NULL;
		}               

	if (hmmioIn != NULL)
	{
	mmioClose(hmmioIn, 0);
		hmmioIn = NULL;
		}
TEMPCLEANUP:
	*phmmioIn = hmmioIn;
	return(nError);
}

int WaveStartDataRead(
					HMMIO *phmmioIn,
					MMCKINFO *pckIn,
					MMCKINFO *pckInRIFF
					)
{
	int nError;
	nError = 0;
	// Do a nice little seek...
	if ((nError = mmioSeek(*phmmioIn, pckInRIFF->dwDataOffset + sizeof(FOURCC), SEEK_SET)) == -1)
		{
		ASSERT(FALSE);
		}
	nError = 0;
	//      Search the input file for for the 'data' chunk.
	pckIn->ckid = mmioFOURCC('d', 'a', 't', 'a');
	if ((nError = mmioDescend(*phmmioIn, pckIn, pckInRIFF, MMIO_FINDCHUNK)) != 0)
		{
		goto ERROR_READING_WAVE;
		}
	goto CLEANUP;
ERROR_READING_WAVE:
CLEANUP:        
	return(nError);
}

int WaveReadFile(
		HMMIO hmmioIn,                          // IN
		UINT cbRead,                            // IN           
		BYTE *pbDest,                           // IN
		MMCKINFO *pckIn,                        // IN.
		UINT *cbActualRead                      // OUT.		
		)
{
	MMIOINFO    mmioinfoIn;         // current status of <hmmioIn>
	int         nError;
	UINT        cT, cbDataIn;
	nError = mmioGetInfo(hmmioIn, &mmioinfoIn, 0);
	if (nError != 0)
		{
		goto ERROR_CANNOT_READ;
		}
	cbDataIn = cbRead;
	if (cbDataIn > pckIn->cksize) 
		cbDataIn = pckIn->cksize;       
	pckIn->cksize -= cbDataIn;
	for (cT = 0; cT < cbDataIn; cT++)
		{
		/* Copy the bytes from the io to the buffer. */
		if (mmioinfoIn.pchNext == mmioinfoIn.pchEndRead)
			{
	    if ((nError = mmioAdvance(hmmioIn, &mmioinfoIn, MMIO_READ)) != 0)
				{
		goto ERROR_CANNOT_READ;
				} 
	    if (mmioinfoIn.pchNext == mmioinfoIn.pchEndRead)
				{
				nError = ER_CORRUPTWAVEFILE;
		goto ERROR_CANNOT_READ;
				}
			}			
		// Actual copy.
		*((BYTE*)pbDest+cT) = *((BYTE*)mmioinfoIn.pchNext)++;                                                                                                   
		}
	if ((nError = mmioSetInfo(hmmioIn, &mmioinfoIn, 0)) != 0)
		{
		goto ERROR_CANNOT_READ;
		}
	*cbActualRead = cbDataIn;
	goto FINISHED_READING;
ERROR_CANNOT_READ:
	*cbActualRead = 0;
FINISHED_READING:
	return(nError);
}

int WaveCloseReadFile(
			HMMIO *phmmio,                          // IN
			WAVEFORMATEX **ppwfxSrc                 // IN
			)
{
	if (*ppwfxSrc != NULL)
		{
		GlobalFree(*ppwfxSrc);
		*ppwfxSrc = NULL;
		}
	if (*phmmio != NULL)
		{
		mmioClose(*phmmio, 0);
		*phmmio = NULL;
		}
	return(0);
}



BOOL PopFileRead (PTSTR pstrFileName)
{
     WAVEFORMATEX  *pwfx;          // Wave format info
     HMMIO         hmmio;          // File handle
     MMCKINFO      mmckinfoData;   // Chunk info
     MMCKINFO      mmckinfoParent; // Parent chunk info
     UINT          dwBytesRead ;
     int           iFileLength;
     short         *pUpConvert = NULL;
     char          Line[80];
     
      pSaveBuffer = realloc(pSaveBuffer,1);
      dwDataLength = 0;
      dwCurrentPos = 0;

          // Open the file.
      if (WaveOpenFile(pstrFileName, &hmmio, &pwfx, &mmckinfoParent) != 0)
      {
         return FALSE;
      }          
      
      // Check for file caveats: WAVE, mono, 16bit, & Size vs Available memory
      // if 8bit allow to pass, and query for permission to upconvert
      if ((pwfx->wFormatTag != WAVE_FORMAT_PCM ) ||
          (pwfx->nChannels != 1) ||
          ((pwfx->wBitsPerSample != 16) && (pwfx->wBitsPerSample != 8))
          )
      {
         MessageBox (NULL, szFormat, szAppName,
                     MB_ICONEXCLAMATION | MB_OK) ;   
         return FALSE;      
      }     
      
      //warn of upconversion to 16bit if 8bit input
      if (pwfx->wBitsPerSample == 8)
      {
      int x;
         x =MessageBox (NULL, szReformat, szAppName,
                          MB_ICONEXCLAMATION | MB_YESNO);
         if (x%2)  
            { 
               return FALSE;
            }
      }
      
       // advance to data
      if (WaveStartDataRead(&hmmio, &mmckinfoData, &mmckinfoParent) != 0)
      {
         // Failure
         WaveCloseReadFile(&hmmio, &pwfx);
         return FALSE;
      }


     // warn user if file is big
     if (mmckinfoData.cksize>1024*1024*64)
     {
     int x;
     char line[80];
         sprintf(line,"Large File size:%d\nContinue Read?",mmckinfoData.cksize);
         x =MessageBox (NULL, line, szAppName,
                          MB_ICONEXCLAMATION | MB_YESNO);
         if (x%2)  
            { 
               return FALSE;
            }
     }

     // Get file size in bytes and allocate memory for read.
     iFileLength = mmckinfoData.cksize; 
     pSaveBuffer = realloc (pSaveBuffer,iFileLength) ;
     
     // halt if could not allocate memory
     if (!pSaveBuffer)
     {
         sprintf(Line,szOutOfMemory,iFileLength);
         MessageBox (NULL, Line, szAppName,
                     MB_ICONWARNING  | MB_OK) ;   
         return FALSE;
     }
     
     // if 8bit input, allocate segment for upconversion
     if (pwfx->wBitsPerSample==8)
     {
        pUpConvert = (short*) malloc(2*iFileLength);
        if (!pUpConvert)    
        {
           sprintf(Line,szOutOfMemory,2*iFileLength);
           MessageBox (NULL, Line, szAppName,
                       MB_ICONWARNING  | MB_OK) ;   
           // Failure
           dwDataLength = 0;
           dwCurrentPos = 0;
           pSaveBuffer = realloc(pSaveBuffer,1);
           return FALSE;
        }
     }
     
     if (WaveReadFile(
                      hmmio,         // file handle
                      iFileLength,   // no. of bytes to read
                      pSaveBuffer,   // destination
                      &mmckinfoData, // file chunk info 
                      &dwBytesRead)) // actual no. of bytes read
    {
       // Handle failure on non-zero return
       WaveCloseReadFile(&hmmio, &pwfx);
       dwDataLength = 0;
       dwCurrentPos = 0;
       pSaveBuffer = realloc(pSaveBuffer,1);
       if (!pUpConvert)    
          free(pUpConvert);
       return FALSE;
    }
    waveform = *pwfx;
    WaveCloseReadFile(&hmmio, &pwfx);      
    if ((UINT)iFileLength == dwBytesRead)
    {
        // if 8bit,upconvert to 16, else continue
        if (pUpConvert)
        {
        int x;
           for (x=0;x<iFileLength;x++)
           {
             pUpConvert[x] =  256*pSaveBuffer[x]-32768;
           }
           // switch over copied buffer
           free(pSaveBuffer);
           pSaveBuffer = (PBYTE) pUpConvert;
           dwDataLength = 2*iFileLength;
           waveform.nBlockAlign    = 2;
           waveform.wBitsPerSample = 16;
           waveform.nAvgBytesPerSec *= 2;
        }
        else
        {
           dwDataLength = iFileLength;
        }
        dwCurrentPos = 0;
        return TRUE;
     }
     else
     {
        dwDataLength = 0;
        dwCurrentPos = 0;
        pSaveBuffer = realloc(pSaveBuffer,1);
        if (!pUpConvert)    
          free(pUpConvert);

        return FALSE;
     }
}

int WaveCreateFile(
			TCHAR*pszFileName,                              // (IN)
			HMMIO *phmmioOut,                               // (OUT)
			WAVEFORMATEX *pwfxDest,                 // (IN)
			MMCKINFO *pckOut,                               // (OUT)
			MMCKINFO *pckOutRIFF                    // (OUT)

			)
{
	    
	int             nError;                         // Return value.
	DWORD           dwFactChunk;            // Contains the actual fact chunk. Garbage until WaveCloseWriteFile.
	MMCKINFO        ckOut1;

	dwFactChunk = (DWORD)-1;
	nError = 0;

	*phmmioOut = mmioOpen(pszFileName, NULL,
			MMIO_ALLOCBUF | MMIO_READWRITE|MMIO_CREATE);

    if (*phmmioOut == NULL)
		{
		nError = ER_CANNOTWRITE;
	goto ERROR_CANNOT_WRITE;    // cannot save WAVE file
		}

    /* Create the output file RIFF chunk of form type 'WAVE'.
     */
    pckOutRIFF->fccType = mmioFOURCC('W', 'A', 'V', 'E');       
	pckOutRIFF->cksize = 0; 
    if ((nError = mmioCreateChunk(*phmmioOut, pckOutRIFF, MMIO_CREATERIFF)) != 0)
		{
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
		}

    /* We are now descended into the 'RIFF' chunk we just created.
     * Now create the 'fmt ' chunk. Since we know the size of this chunk,
     * specify it in the MMCKINFO structure so MMIO doesn't have to seek
     * back and set the chunk size after ascending from the chunk.
     */
    pckOut->ckid = mmioFOURCC('f', 'm', 't', ' ');
    pckOut->cksize = sizeof(PCMWAVEFORMAT);   // we know the size of this ck.
    if ((nError = mmioCreateChunk(*phmmioOut, pckOut, 0)) != 0)
		{
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
		}

    /* Write the PCMWAVEFORMAT structure to the 'fmt ' chunk if its that type. */
    if (pwfxDest->wFormatTag == WAVE_FORMAT_PCM)
		{
	    if (mmioWrite(*phmmioOut, (HPSTR) pwfxDest, sizeof(PCMWAVEFORMAT))
		!= sizeof(PCMWAVEFORMAT))
			{
			nError = ER_CANNOTWRITE;
		goto ERROR_CANNOT_WRITE;    // cannot write file, probably
			}
		}
	
	else 
		{
		// Write the variable length size.
		if ((UINT)mmioWrite(*phmmioOut, (HPSTR) pwfxDest, sizeof(*pwfxDest)+pwfxDest->cbSize)
		!= (sizeof(*pwfxDest)+pwfxDest->cbSize))
			{
			nError = ER_CANNOTWRITE;
		goto ERROR_CANNOT_WRITE;    // cannot write file, probably
			} 

		}  

    /* Ascend out of the 'fmt ' chunk, back into the 'RIFF' chunk.
     */
    if ((nError = mmioAscend(*phmmioOut, pckOut, 0)) != 0)
		{
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
		}

	// Now create the fact chunk, not required for PCM but nice to have.  This is filled
	// in when the close routine is called.
	ckOut1.ckid = mmioFOURCC('f', 'a', 'c', 't');
	ckOut1.cksize = 0;
    if ((nError = mmioCreateChunk(*phmmioOut, &ckOut1, 0)) != 0)
		{
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
		}

	if (mmioWrite(*phmmioOut, (HPSTR)&dwFactChunk, sizeof(dwFactChunk)) != sizeof(dwFactChunk))
		{
		nError = ER_CANNOTWRITE;
		goto ERROR_CANNOT_WRITE;
		}

	// Now ascend out of the fact chunk...
    if ((nError = mmioAscend(*phmmioOut, &ckOut1, 0)) != 0)
		{
	nError = ER_CANNOTWRITE;        // cannot write file, probably
		goto ERROR_CANNOT_WRITE;
		}

	 
   
	goto DONE_CREATE;

ERROR_CANNOT_WRITE:
	// Maybe delete the half-written file?  Ah forget it for now, its good to leave the
	// file there for debugging...

DONE_CREATE:
	return(nError);

}

/*      This routine has to be called before any data is written to the wave output file, via wavewritefile.  This
	sets up the data to write, and creates the data chunk.
*/

int WaveStartDataWrite(
				HMMIO *phmmioOut,                               // (IN)
				MMCKINFO *pckOut,                               // (IN)
				MMIOINFO *pmmioinfoOut                  // (OUT)
				)
{
	int             nError;
	nError = 0;
 /* Create the 'data' chunk that holds the waveform samples.  */
    pckOut->ckid = mmioFOURCC('d', 'a', 't', 'a');
	pckOut->cksize = 0;
    if ((nError = mmioCreateChunk(*phmmioOut, pckOut, 0)) != 0)
	{
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
	}
	if ((nError = mmioGetInfo(*phmmioOut, pmmioinfoOut, 0)) != 0)
    {
	goto ERROR_CANNOT_WRITE;
	}
	goto CLEANUP;
ERROR_CANNOT_WRITE:     
CLEANUP:
	return(nError);
}

int WaveWriteFile(
		HMMIO hmmioOut,                         // (IN)
		UINT cbWrite,                           // (IN)
		BYTE *pbSrc,                            // (IN)
		MMCKINFO *pckOut,                       // (IN)
		UINT *cbActualWrite,            // (OUT)
		MMIOINFO *pmmioinfoOut          // (IN)
		)
{
	int   nError;
	UINT  cT;
	nError = 0;
	*cbActualWrite = 0;
	for (cT=0; cT < cbWrite; cT++)
		{       
		if (pmmioinfoOut->pchNext == pmmioinfoOut->pchEndWrite)
		{
		pmmioinfoOut->dwFlags |= MMIO_DIRTY;
		if ((nError = mmioAdvance(hmmioOut, pmmioinfoOut, MMIO_WRITE)) != 0)
				{
		    goto ERROR_CANNOT_WRITE;
				}
		}
		*((BYTE*)pmmioinfoOut->pchNext)++ = *((BYTE*)pbSrc+cT);
		(*cbActualWrite)++;
		}	
ERROR_CANNOT_WRITE:
	return(nError);
}


int WaveCloseWriteFile(
			HMMIO *phmmioOut,               // (IN)
			MMCKINFO *pckOut,               // (IN)
			MMCKINFO *pckOutRIFF,   // (IN)
			MMIOINFO *pmmioinfoOut, // (IN)
			DWORD cSamples                  // (IN)
			)
{
	int nError;                         				
	nError = 0;
	if (*phmmioOut == NULL)
		return(0);
	pmmioinfoOut->dwFlags |= MMIO_DIRTY;
    if ((nError = mmioSetInfo(*phmmioOut, pmmioinfoOut, 0)) != 0)
		{
		// cannot flush, probably...
	goto ERROR_CANNOT_WRITE;                                
		}
	 /* Ascend the output file out of the 'data' chunk -- this will cause
     * the chunk size of the 'data' chunk to be written.
     */
    if ((nError = mmioAscend(*phmmioOut, pckOut, 0)) != 0)
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
	// Do this here instead...
	if ((nError = mmioAscend(*phmmioOut, pckOutRIFF, 0)) != 0)
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
	nError = mmioSeek(*phmmioOut, 0, SEEK_SET);
	if ((nError = (int)mmioDescend(*phmmioOut, pckOutRIFF, NULL, 0)) != 0)
		{
		goto ERROR_CANNOT_WRITE;
		}
	nError = 0;
	pckOut->ckid = mmioFOURCC('f', 'a', 'c', 't');
	if ((nError = mmioDescend(*phmmioOut, pckOut, pckOutRIFF, MMIO_FINDCHUNK)) == 0)
		{
		// If it didn't fail, write the fact chunk out, if it failed, not critical, just
		// assert (below).
	    nError = mmioWrite(*phmmioOut, (HPSTR)&cSamples, sizeof(DWORD));
		nError = mmioAscend(*phmmioOut, pckOut, 0); 
		nError = 0;
		}
	else 
		{
		nError = 0;
		ASSERT(FALSE);
		}
// CANTWRITEFACT:
    /* Ascend the output file out of the 'RIFF' chunk -- this will cause
     * the chunk size of the 'RIFF' chunk to be written.
     */
    if ((nError = mmioAscend(*phmmioOut, pckOutRIFF, 0)) != 0)
	goto ERROR_CANNOT_WRITE;    // cannot write file, probably
ERROR_CANNOT_WRITE:
	if (*phmmioOut != NULL)
		{
	mmioClose(*phmmioOut, 0);
		*phmmioOut = NULL;
		}
	return(nError);
}


BOOL PopFileWrite (PTSTR pstrFileName)
{
WAVEFORMATEX  wfx;            // Wave format info
HMMIO         hmmio;          // File handle
MMCKINFO      mmckinfoData;   // Chunk info
MMCKINFO      mmckinfoParent; // Parent chunk (file) info
MMIOINFO      mmioinfo;
UINT          dwBytesWritten;
   wfx = waveform;
   if (WaveCreateFile(pstrFileName, &hmmio, &wfx, 
        &mmckinfoData, &mmckinfoParent))
   {
       return FALSE;
   }
   if (WaveStartDataWrite(&hmmio, &mmckinfoData, &mmioinfo))
   {
      CloseHandle (hmmio) ;     
      return FALSE;
   }
   if (WaveWriteFile(hmmio, dwDataLength, pSaveBuffer, &mmckinfoData,
                     &dwBytesWritten, &mmioinfo))
   {
      CloseHandle (hmmio) ;     
      return FALSE;
   }
   WaveCloseWriteFile(&hmmio, &mmckinfoData, 
                      &mmckinfoParent, &mmioinfo,
                      dwBytesWritten / (wfx.wBitsPerSample / 8));
   return TRUE;
}




BOOL PopFileOpenDlg (HWND hwnd, PTSTR pstrFileName, PTSTR pstrTitleName)
{
     ofn.hwndOwner         = hwnd ;
     ofn.lpstrFile         = pstrFileName ;
     ofn.lpstrFileTitle    = pstrTitleName ;
     ofn.Flags             = OFN_HIDEREADONLY | OFN_CREATEPROMPT ;
 
     return GetOpenFileName (&ofn) ;
}

BOOL PopFileSaveDlg (HWND hwnd, PTSTR pstrFileName, PTSTR pstrTitleName)
{
     ofn.hwndOwner         = hwnd ;
     ofn.lpstrFile         = pstrFileName ;
     ofn.lpstrFileTitle    = pstrTitleName ;
     ofn.Flags             = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT ;
     
     return GetSaveFileName (&ofn) ;
}

void InitPoints()
{
int GMedian;
int x;
   GMedian = GRY - (int)((double) GRY * (Average_Frequency-Lowest_Frequency)/(Frequency_Range));
   for (x=0;x<GRAPHPOINTCOUNT;x++)
   {
      GraphPoints[x].x = (GRX-1) * x / GRAPHPOINTCOUNT;
      GraphPoints[x].y = GMedian;
   }
}



void PlotPoints(HDC hdc)
{
int iDrawMode;
   //save old drawing mode
   iDrawMode = GetROP2 (hdc) ;
   // set XOR DRAW
   SetROP2 (hdc, R2_XORPEN) ;
   MoveToEx (hdc, GraphPoints[0].x, GraphPoints[0].y, NULL) ;
   PolylineTo (hdc, GraphPoints + 1, GRAPHPOINTCOUNT-1) ;
   // restore old drawing mode
   SetROP2 (hdc, iDrawMode) ;
}


void UpdateGraph (double Frequency)
{
HPEN   hPen1;
HDC    hdc ;
int    x;  
int    y;   
   if (SampleValid==FALSE)
       return;
   hdc = GetDC (hwndGraph) ;
   hPen1 = CreatePen (PS_SOLID,  1, RGB (255,255,255)) ;
   SelectObject (hdc, hPen1) ;
   PlotPoints(hdc);
   
   // Scroll data left
   for (x=1;x<GRAPHPOINTCOUNT;x++)
   {
      GraphPoints[x-1].y = GraphPoints[x].y;
   }
   y = GRY - (int) ((double) GRY * (Frequency-Lowest_Frequency)/(Frequency_Range));
   if (y<1) y = 1;
   if (y>GRY-1) y = GRY-1;
   GraphPoints[GRAPHPOINTCOUNT-1].y = y;   
   PlotPoints(hdc);
   DeleteObject (SelectObject (hdc, hPen1)) ;
   ReleaseDC (hwndGraph, hdc) ;
}

void RepaintWindow()
{
     static HDC          hdc ;
     static HBRUSH       hBrush1,hBrush2 ;
     static HPEN         hPen1, hPen2, hPen3;
     int    GMedian;
     int    DeltaHi;
     int    DeltaLow;

          hdc = GetDC (hwndGraph) ;
          hBrush1 = CreateSolidBrush (RGB(64,64,64)) ;
          SelectObject (hdc, hBrush1) ;     
          Rectangle   (hdc, 0, 0, GRX, GRY) ; 
          hBrush2 = CreateSolidBrush (RGB(255,0,0)) ;
          hPen1 = CreatePen (PS_SOLID,  1, RGB (255, 255, 255)) ;
          hPen2 = CreatePen (PS_SOLID,  3, RGB (  0, 255,   0)) ;
          hPen3 = CreatePen (PS_SOLID,  1, RGB (255,   0,   0)) ;
          GMedian = GRY - (int)((double) GRY * (Average_Frequency-Lowest_Frequency)/(Frequency_Range));
          SelectObject (hdc, hPen2) ;
          MoveToEx (hdc,1,GMedian,NULL);
          LineTo   (hdc,GRX-3,GMedian);
          SelectObject (hdc, hPen3) ;
          SelectObject (hdc, hBrush2) ;     
          DeltaHi = (int)((double) GRY*
                                    ((1.0-Frequency_Threshold)*Frequency_Range)/Frequency_Range);
          DeltaLow= (int)((double) GRY*
                                    ((    Frequency_Threshold)*Frequency_Range)/Frequency_Range);
          Rectangle(hdc,1,1,GRX-1,GRY-DeltaHi);
          Rectangle(hdc,1,GRY-DeltaLow,GRX-1,GRY-1);
          SelectObject (hdc, hPen1) ;
          PlotPoints(hdc);
          DeleteObject (SelectObject (hdc, hBrush1)) ;
          DeleteObject (SelectObject (hdc, hBrush2)) ;
          DeleteObject (SelectObject (hdc, hPen1)) ;
          DeleteObject (SelectObject (hdc, hPen2)) ;
          DeleteObject (SelectObject (hdc, hPen3)) ;
          ReleaseDC (hwndGraph, hdc) ;
}


void Announce(char  Operation[],
              int   Current_Position,
              int   Last_Position)
{
int Position;
int Value1;
int Value2;
int Minutes1;
int Seconds1;
int Minutes2;
int Seconds2;
char  Line[80];
    // if no data in memory,
    // then waveform is invalid, avoid divide by zero
    if (dwDataLength == 0)
    { 
      sprintf(Line,"%s",Operation);
    }
    else
    if (Last_Position != 0)
    {
       Value1   = Current_Position / waveform.nAvgBytesPerSec;
       Minutes1 = Value1 / 60;
       Seconds1 = Value1 % 60;
       
       Value2  = Last_Position / waveform.nAvgBytesPerSec;
       Minutes2 = Value2 / 60;
       Seconds2 = Value2 % 60;
       
       sprintf(Line,"%s %d:%02d/%d:%02d",
               Operation,
               Minutes1,Seconds1,
               Minutes2,Seconds2);

       SetScrollRange (hwndPlayback, SB_CTL, 0, PLAYBACK_RESOLUTION, TRUE) ;
       // use floating point to avoid integer truncation in very large audio files
       Position = (int) ((double) PLAYBACK_RESOLUTION* (double) Current_Position/ (double)Last_Position);
       SetScrollPos   (hwndPlayback, SB_CTL, Position, TRUE) ;
    }
    else
    {
       Value1   = Current_Position / waveform.nAvgBytesPerSec;
       Minutes1 = Value1 / 60;
       Seconds1 = Value1 % 60;
       
       sprintf(Line,"%s %d:%02d",
               Operation,
               Minutes1,Seconds1);
    }
    SetWindowText(hwndMessage,Line);
}

void SetDialogState(
                      HWND  hwnd,
                      HWND  hwndMessage,
                      HWND  hwndPlayback,
                      BOOL  Play,
                      BOOL  Pause,
                      BOOL  Stop,
                      BOOL  Record,
                      int   Focus,
                      int   dwActualPos,
                      int   dwDataLength)
{
   EnableWindow (GetDlgItem (hwnd, IDC_PLAY),   Play) ;
   EnableWindow (GetDlgItem (hwnd, IDC_PAUSE),  Pause) ;
   EnableWindow (GetDlgItem (hwnd, IDC_STOP),   Stop)  ;
   EnableWindow (GetDlgItem (hwnd, IDC_RECORD), Record) ;
   SetFocus (GetDlgItem (hwnd, Focus)) ;
   
   // disable playback scroll unless play is valid
   if ((Play) || (Pause))
   {
      ShowWindow (hwndPlayback, SW_SHOWNORMAL);
   }
   else
   {
      ShowWindow (hwndPlayback, SW_HIDE);
   }
   
   // announce how much data we have
   if (dwDataLength != 0)
   {
      Announce("Ready",dwActualPos,dwDataLength);
   }
   else
   {
      SetWindowText(hwndMessage,"Ready");
   }
   RepaintWindow();
}

void MenuReset(
               BOOL  Play,
               BOOL  Pause,
               BOOL  Stop,
               BOOL  Record,
               int   Focus)
{
    // reclaim the save buffer
    if (dwDataLength > 0)
       pSaveBuffer = realloc (pSaveBuffer, 1) ;
          
    // init sampling vars
    dwCurrentPos = 0;
    dwDataLength = 0 ;
    bNeedSave = FALSE ;
               
    // Enable and disable Buttons
    SetDialogState(hwndModeless,hwndMessage,hwndPlayback,Play,Pause,Stop,Record,Focus,dwCurrentPos,dwDataLength);
}


DWORD MidiOutMessage (HMIDIOUT hMidi, int iStatus, int iChannel,
                                      int iData1,  int iData2)
{
     DWORD dwMessage = iStatus | iChannel | (iData1 << 8) | (iData2 << 16) ;
     
     return midiOutShortMsg (hMidi, dwMessage) ;
}




double CalcFrequencyMagnitude(double  Frequency,
                              short   Data[],
                              int     NumSamples)
{
int k;
double R0, I0, cos1, cos2, cos3, sin1, sin2, sin3, beta, theta;

   theta = 6.28318530717959 * Frequency / waveform.nSamplesPerSec;
   sin1 = sin(-2.0 * theta);
   sin2 = sin(     - theta);
   cos1 = cos(-2.0 * theta);
   cos2 = cos(     - theta);
   
   beta = 2.0 * cos2;
   
   R0 = 0.0; 
   I0 = 0.0;
   
   for (k=0;k<NumSamples;k++)
   {
      sin3 = beta * sin2 - sin1;
      sin1 = sin2;
      sin2 = sin3;
      cos3 = beta * cos2 - cos1;
      cos1 = cos2;
      cos2 = cos3;
      
      R0 = R0 + (double)Data[k] * cos3;
      I0 = I0 + (double)Data[k] * sin3;
   }
   return sqrt(R0*R0 + I0*I0);
}


double GetAverageFrequency(double rx[])
{
double ry[41];
int    x,y;
double r,t,t2;

   for (x=0;x<=7;x++)
   {
      if (rx[x+1] != rx[x])
      {
         r = (rx[x+1] - rx[x]) / 5.0;
         for (y=0;y<=4;y++)
         {
            ry[x*5+y] = rx[x] + (double) y * r;
         }    
      }
      else
      {
         for (y=0;y<=4;y++)
         {
            ry[x*5+y] = rx[x];
         }
      }
   }
   ry[40] = rx[8];
   t = 0.0;
   for (x=0;x<=40;x++)
   {
      t += ry[x];
   }
   t  = t/2.0;
   t2 = 0.0;
   x=0;
   do 
   {
     t2 = t2+ry[x];
     x++;
   } while (t2 < t);
   return (double) (x-1)/2.0;
}


void Analyze(PBYTE data,
             int   ByteCount)
{
int x;
int y;
double Frequency;
double Volume;
int BlockCount;
int WordCount = ByteCount/2;
int WordsLeft;
double Spectrum[9];


   SampleValid = FALSE;
   Average_Volume = 0.0;
   
   // step across data by word
   for (x=0;x<WordCount;x++)
   {
       Average_Volume += (double) (int)((short*)(data)[x]);
   }
   Average_Volume /= (double) WordCount;
   BlockCount = 0;
   Lowest_Frequency  = 20.0;
   Highest_Frequency = 0.0;
   Average_Frequency = 0.0;
   for (x=0;x<WordCount;x+=waveform.nSamplesPerSec)
   {      
      Volume = 0.0;
      if ((UINT)WordCount-x-y<waveform.nSamplesPerSec)
          WordsLeft = WordCount-x-y;
      else
          WordsLeft = waveform.nSamplesPerSec;
      for (y=0;y<WordsLeft;y++)
      {
          Volume += (double) (int)((short*)(data)[x+y]) ;
      }
      Volume /= (double) WordsLeft;
      if (Volume > Average_Volume)
      {
         Spectrum[0] = CalcFrequencyMagnitude(8.0,&((short*) data)[x],WordsLeft);
         Spectrum[1] = CalcFrequencyMagnitude(8.5,&((short*) data)[x],WordsLeft);
         Spectrum[2] = CalcFrequencyMagnitude(9.0,&((short*) data)[x],WordsLeft);
         Spectrum[3] = CalcFrequencyMagnitude(9.5,&((short*) data)[x],WordsLeft);
         Spectrum[4] = CalcFrequencyMagnitude(10.0,&((short*) data)[x],WordsLeft);
         Spectrum[5] = CalcFrequencyMagnitude(10.5,&((short*) data)[x],WordsLeft);
         Spectrum[6] = CalcFrequencyMagnitude(11.0,&((short*) data)[x],WordsLeft);
         Spectrum[7] = CalcFrequencyMagnitude(11.5,&((short*) data)[x],WordsLeft);
         Spectrum[8] = CalcFrequencyMagnitude(12.0,&((short*) data)[x],WordsLeft);
         Frequency = GetAverageFrequency(Spectrum);
         Average_Frequency += Frequency;

         if (Frequency > Highest_Frequency)
            Highest_Frequency = Frequency;
         if (Frequency < Lowest_Frequency)
            Lowest_Frequency = Frequency;
         BlockCount++;
      }
   } 
   Average_Frequency /= (double) BlockCount;
   if (Highest_Frequency == Lowest_Frequency)
      Frequency_Range = 1.0;
   else
      Frequency_Range = Highest_Frequency - Lowest_Frequency;
   if (Frequency_Range < 1.0)
      Frequency_Range = 1.0;
   SampleValid = TRUE;
   // force window to repaint and null out data
   InitPoints();
   SendMessage (hwndModeless, WM_PAINT, 0, 0) ;
}

BOOL CALLBACK DlgProc (HWND hDlg, UINT message, 
                       WPARAM wParam, LPARAM lParam)
{
     switch (message)
     {     
     case WM_INITDIALOG :          
          return TRUE ;          
     case WM_COMMAND :          
          switch (LOWORD (wParam))          
          {
          case IDOK :          
          case IDCANCEL :
               EndDialog (hDlg, 0) ; 
               return TRUE ;          
          }
     break ;
     }
     return FALSE ;
}




void QueueWaveOut(HWAVEOUT  hWaveOut,
                  PWAVEHDR  pWaveHdr,
                  PBYTE     pSaveBuffer,
                  int     *dwCurrentPos,
                  int      dwDataLength)
{
int    WordsLeft;
   int    x;
   int    y;
   double Volume;
   double Spectrum[9];
   double Frequency;

   // save starting position in words
   // for fourier calculation
   x = *dwCurrentPos/2;
   pWaveHdr->lpData          = (char*)&pSaveBuffer[*dwCurrentPos] ;
   // send as multi-block operation, instead of a single burst
   // this will allow us to intercept and change position
   if ((UINT)(dwDataLength - *dwCurrentPos) < waveform.nAvgBytesPerSec) 
   {
      pWaveHdr->dwBufferLength  = dwDataLength - *dwCurrentPos ;
      *dwCurrentPos += dwDataLength - *dwCurrentPos;
   }
   else
   {
      pWaveHdr->dwBufferLength  =  waveform.nAvgBytesPerSec;
      *dwCurrentPos             += waveform.nAvgBytesPerSec;
   }
   pWaveHdr->dwBytesRecorded = 0 ;
   pWaveHdr->dwUser          = 0 ;
   pWaveHdr->dwFlags         = 0;
   pWaveHdr->dwLoops         = 1 ;
   pWaveHdr->lpNext          = NULL ;
   pWaveHdr->reserved        = 0 ;


   // if full analysis successful, do realtime interpolation
   // over sound bite interval
   if (SampleValid==TRUE)
   {  
      WordsLeft = pWaveHdr->dwBufferLength/2;
      Volume = 0.0;
      for (y=0;y<WordsLeft;y++)
      {
          Volume += (double) (int)((short*)(pSaveBuffer)[x+y]) ;
      }
      Volume /= (double) WordsLeft;
      // perform fourier only on sound bites that exceed average volume
      // this supresses ambient noise in spectrum calculations
      if (Volume > Average_Volume)
      {      
         Spectrum[0] = CalcFrequencyMagnitude(8.0,&((short*)  pSaveBuffer)[x],WordsLeft);
         Spectrum[1] = CalcFrequencyMagnitude(8.5,&((short*)  pSaveBuffer)[x],WordsLeft);
         Spectrum[2] = CalcFrequencyMagnitude(9.0,&((short*)  pSaveBuffer)[x],WordsLeft);
         Spectrum[3] = CalcFrequencyMagnitude(9.5,&((short*)  pSaveBuffer)[x],WordsLeft);
         Spectrum[4] = CalcFrequencyMagnitude(10.0,&((short*) pSaveBuffer)[x],WordsLeft);
         Spectrum[5] = CalcFrequencyMagnitude(10.5,&((short*) pSaveBuffer)[x],WordsLeft);
         Spectrum[6] = CalcFrequencyMagnitude(11.0,&((short*) pSaveBuffer)[x],WordsLeft);
         Spectrum[7] = CalcFrequencyMagnitude(11.5,&((short*) pSaveBuffer)[x],WordsLeft);
         Spectrum[8] = CalcFrequencyMagnitude(12.0,&((short*) pSaveBuffer)[x],WordsLeft);
         Frequency = GetAverageFrequency(Spectrum);
         UpdateGraph(Frequency);
      }
      else
      {
         UpdateGraph(Average_Frequency);
      }
   }

   // Prepare and write     
   waveOutPrepareHeader (hWaveOut, pWaveHdr, sizeof (WAVEHDR)) ;
   waveOutWrite (hWaveOut, pWaveHdr, sizeof (WAVEHDR)) ;
   
}


LRESULT CALLBACK MainDlg (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
     static struct
     {
          int iDur ;
          int iNote [2] ;
     }
     noteseq [] = { 210, 57, 69,
                    110, 55, 67, 1190, 57, 69,  320, -1, -1,  320, 52, 64,
                    320, 53, 65,  320, 49, 61,  940, 50, 62, 300, -1, -1 } ;
     
     static HMIDIOUT    hMidiOut ;
     static int         iIndex ;
     static HMENU       hMenu ;
     static HINSTANCE   hInstance ;
     static PAINTSTRUCT ps;
            char        Line[255];

     
     static BOOL         bEnding, bTerminating, bPaused ;
     
     static HWAVEIN      hWaveIn ;
     static HWAVEOUT     hWaveOut ;
     static PBYTE        pBuffer1, pBuffer2, pNewBuffer ;
     static PWAVEHDR     pWaveHdr1, pWaveHdr2 ;

     static RECT         rect ;                            
            BOOL         Error;
            WAVEINCAPS   wic;
            int          Value;
     
     switch (message)
     {
     case WM_INITDIALOG:
          hwndModeless = hwnd;
          bNeedSave = FALSE;
          PopFileInitialize (hwnd) ;     

          hInstance = (HINSTANCE) GetWindowLong (hwnd, GWL_HINSTANCE) ;
          hMenu     = LoadMenu (hInstance, szAppName) ;
          SetMenu (hwnd, hMenu) ;
          DrawMenuBar (hwnd) ;
          
          hwndBuddy    = GetDlgItem (hwnd, IDC_BUDDY);
          hwndSpinCtl  = GetDlgItem (hwnd, IDC_SPINCTL);
          hwndGraph    = GetDlgItem (hwnd, IDC_GRAPH);
          hwndMessage  = GetDlgItem (hwnd, IDC_TEXTOUT) ;
          hwndPlayback = GetDlgItem (hwnd, IDC_PLAYBACK);

          dwDataLength = 0;
          dwCurrentPos = 0;
          
               // Allocate memory for wave header
          
          pWaveHdr1 = malloc (sizeof (WAVEHDR)) ;
          pWaveHdr2 = malloc (sizeof (WAVEHDR)) ;
          
               // Allocate memory for save buffer
          pSaveBuffer = malloc (1) ;
          
          // setup temp value, overridden with file/load
          waveform.nSamplesPerSec = DEFAULT_SAMPLE_RATE;

          GetClientRect (hwndGraph, &rect) ;
          GRX = rect.right;
          GRY = rect.bottom;
          
          Average_Frequency = 10.0;
          Highest_Frequency = 20.0;
          Lowest_Frequency  =  0.0;
          Frequency_Range   = 20.0;
          Frequency_Threshold = 0.1;
          SampleValid = FALSE;
          InitPoints();
          
          
          // Set the buddy window.
          SendMessage (hwndSpinCtl, UDM_SETBUDDY, (LONG)hwndBuddy, 0L);
          // Set the range.
          SendMessage (hwndSpinCtl, UDM_SETRANGE, 0L, 
                       MAKELONG (50, 1));
          // Set the default value in the edit control.
          SetDlgItemInt (hwnd, IDC_BUDDY,(int)(100.0*Frequency_Threshold), FALSE);

          SetDialogState(hwnd,hwndMessage,hwndPlayback,FALSE,FALSE,FALSE,TRUE,IDC_RECORD,dwCurrentPos,dwDataLength);

          // activate auto-play if command line present
          if ((CmdLine) && (CmdLine[0])) 
          {
             if (!PopFileRead (CmdLine))
                 {
                        OkMessage (hwnd, TEXT ("Could not read file %s!"),
                                    CmdLine) ;
                 }
                 else
                 {
                     sprintf(szFileName,"%s",CmdLine);
                     sprintf(szTitleName,"%s",CmdLine);

                     // analyze file read
                     SetWindowText(hwndMessage,TEXT("Analyzing File"));
                     Analyze(pSaveBuffer,dwDataLength);
                     
                     // set button states, but preserve loaded content
                     SetDialogState(hwnd,hwndMessage,hwndPlayback,TRUE,FALSE,TRUE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength);

                    DoCaption (hwnd, szTitleName) ;
                    SendMessage (hwnd, WM_COMMAND, IDC_PLAY, 0L) ;
                 }
          }
          else
          {
               szFileName[0]  = '\0' ;
               szTitleName[0] = '\0' ;
               DoCaption (hwnd, szTitleName) ;
          }

          return TRUE;  
     case WM_PAINT:
          BeginPaint (hwnd, &ps) ;     
          EndPaint (hwnd, &ps) ;
          RepaintWindow();
          return TRUE;
     case WM_COMMAND:
          switch (LOWORD (wParam))
          {
          case IDM_FILE_NEW:
               // continue only if idle
               if ((bPlaying) || (bRecording))
               {
                    MessageBox (hwnd, szBusyError, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
                    return 0;
                
               }
          
               // handle unsaved data
               if (bNeedSave && IDCANCEL == AskAboutSave (hwnd, szTitleName))
                    return 0 ;
              
               szFileName[0]  = '\0' ;
               szTitleName[0] = '\0' ;
               DoCaption (hwnd, szTitleName) ;
               
               // reset in-memory parameters and button states
               MenuReset(FALSE,FALSE,FALSE,TRUE,IDC_RECORD);
               SampleValid = FALSE;
               return 0 ;
               
          case IDM_FILE_OPEN:
               // continue only if idle
               if ((bPlaying) || (bRecording))
               {
                    MessageBox (hwnd, szBusyError, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
                    return 0;
               }
               
               // handle unsaved data
               if (bNeedSave && IDCANCEL == AskAboutSave (hwnd, szTitleName))
                    return 0 ;

               if (PopFileOpenDlg (hwnd, szFileName, szTitleName))
               {
                 SetWindowText(hwndMessage,TEXT("Reading/Analyzing File..."));
                 if (!PopFileRead (szFileName))
                 {
                        OkMessage (hwnd, TEXT ("Could not read file %s!"),
                                    szTitleName) ;
                        szFileName[0]  = '\0' ;
                        szTitleName[0] = '\0' ;
                        // reset in-memory parameters and button states
                        MenuReset(FALSE,FALSE,FALSE,TRUE,IDC_PLAY);

                 }
                 else
                 {
                     // analyze file read
                     Analyze(pSaveBuffer,dwDataLength);                     
                     
                     // set button states, but preserve loaded content
                     SetDialogState(hwnd,hwndMessage,hwndPlayback,TRUE,FALSE,TRUE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength);
                  }
               }
               bNeedSave = FALSE;
               DoCaption (hwnd, szTitleName) ;
               return 0 ;
               
          case IDM_FILE_SAVE_AS:
          case IDM_FILE_SAVE:
               if (!dwDataLength)
               {
                    MessageBox (hwnd, szNoData, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
                    return 0;
               }
               
               // continue only if idle
               if ((bPlaying) || (bRecording))
               {
                    MessageBox (hwnd, szBusyError, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
                    return 0;
               }
               //  handle case where File-save has no filename
               if ((szFileName[0]) && (LOWORD (wParam) == IDM_FILE_SAVE))
               {
                    if (PopFileWrite (szFileName))
                    {
                         bNeedSave = FALSE ;
                         return 1 ;
                    }
                    else
                    {
                         OkMessage (hwnd, TEXT ("Could not write file %s"),
                                    szTitleName) ;
                         return 0 ;
                    }
               }
               else
               {
                   if (PopFileSaveDlg (hwnd, szFileName, szTitleName))
                   {
                       DoCaption (hwnd, szTitleName) ;                    
                       if (PopFileWrite (szFileName))
                       {
                           bNeedSave = FALSE ;
                           return 1 ;
                       }
                       else
                       {
                            OkMessage (hwnd, TEXT ("Could not write file %s"),
                                       szTitleName) ;
                            return 0 ;
                       }
                   }
               }
               return 0 ;
          case IDM_FILE_EXIT:
               SendMessage (hwnd, WM_CLOSE, 0, 0) ;
               return 0 ;
          case IDM_VIEW_STATISTICS:
               {
                  if (SampleValid)
                     sprintf(Line,"Average Volume:       %d\nAverage Frequency:  %0.2f\nLowest Frequency:     %0.2f\nHighest Frequency:   %0.2f",
                            (int) Average_Volume,
                            Average_Frequency,
                            Lowest_Frequency,
                            Highest_Frequency);
                  else 
                     sprintf(Line,"No Data Analyzed");
                  MessageBox (hwnd, Line, szAppName,
                              MB_ICONEXCLAMATION | MB_OK) ;
                  return 0;          
               }
          case IDM_HELP_WHATIS:
               DialogBox (hInstance, TEXT ("WHATIS"), hwnd, DlgProc) ;
               return 0;
          case IDM_HELP_HOWTO:       
               DialogBox (hInstance, TEXT ("HOWTO"), hwnd, DlgProc);
               return 0 ;
          case IDM_HELP_ABOUT:
               if (!midiOutOpen (&hMidiOut, MIDIMAPPER, 0, 0, 0))
               {
                  iIndex = 0;
                  MidiOutMessage (hMidiOut, 0xC0,  0, 19, 0) ;
                  SendMessage (hwnd, WM_TIMER, ID_TIMER, 0) ;
               }      
               DialogBox (hInstance, TEXT ("About"), hwnd, DlgProc) ;
               return 0 ;
          case IDC_RECORD:
               if ((bRecording) || (bPlaying))
               {
                   return TRUE;
               }
               
               // eliminate old filename
               szFileName[0]  = '\0' ;
               szTitleName[0] = '\0' ;

                              
                    // Open waveform audio for input
               
               waveform.wFormatTag      = WAVE_FORMAT_PCM ;
               waveform.nChannels       = 1 ;
               waveform.nSamplesPerSec  = DEFAULT_SAMPLE_RATE ;
               waveform.nBlockAlign     = 2 ;
               waveform.wBitsPerSample  = 16 ;
               waveform.nAvgBytesPerSec = waveform.wBitsPerSample/8*waveform.nSamplesPerSec;
               waveform.cbSize          = 0 ;
               
               // check capabilities of device for compatibility
               if (!waveInGetDevCaps(WAVE_MAPPER,&wic,sizeof(wic)))
               {
                  // accept the lowest fidelty format to conserve space
                  if ((wic.dwFormats&WAVE_FORMAT_1M08) || // 11.025 kHz, mono, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_1M16) || // 11.025 kHz, mono, 16-bit
                      (wic.dwFormats&WAVE_FORMAT_1S08) || // 11.025 kHz, stereo, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_1S16)) // 11.025 kHz, stereo, 16-bit
                  {
                     waveform.nSamplesPerSec  = 11025 ;
                     waveform.nAvgBytesPerSec = waveform.wBitsPerSample/8*waveform.nSamplesPerSec;
                  }
                  else
                  if ((wic.dwFormats&WAVE_FORMAT_2M08) || // 22.05 kHz, mono, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_2M16) || // 22.05 kHz, mono, 16-bit
                      (wic.dwFormats&WAVE_FORMAT_2S08) || // 22.05 kHz, stereo, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_2S16)) // 22.05 kHz, stereo, 16-bit
                  {
                     waveform.nSamplesPerSec  = 22050 ;
                     waveform.nAvgBytesPerSec = waveform.wBitsPerSample/8*waveform.nSamplesPerSec;
                  }
                  else
                  if ((wic.dwFormats&WAVE_FORMAT_4M08) || // 44.1 kHz, mono, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_4M16) || // 44.1 kHz, mono, 16-bit
                      (wic.dwFormats&WAVE_FORMAT_4S08) || // 44.1 kHz, stereo, 8-bit
                      (wic.dwFormats&WAVE_FORMAT_4S16)) // 44.1 kHz, stereo, 16-bit
                  {
                     waveform.nSamplesPerSec  = 44100 ;
                     waveform.nAvgBytesPerSec = waveform.wBitsPerSample/8*waveform.nSamplesPerSec;
                  }
                  // if no match, stick with default set above
              }

                    // Allocate buffer memory 
                    // (record in 10second chunks due to inefficiency in realloc)
                    // 
#define RATE_MULTIPLIER 10
               pBuffer1 = malloc (RATE_MULTIPLIER*waveform.nAvgBytesPerSec) ;
               pBuffer2 = malloc (RATE_MULTIPLIER*waveform.nAvgBytesPerSec) ;
               
               if (!pBuffer1 || !pBuffer2)
               {
                    if (pBuffer1) free (pBuffer1) ;
                    if (pBuffer2) free (pBuffer2) ;

                    MessageBeep (MB_ICONEXCLAMATION) ;
                    sprintf(Line,szOutOfMemory,RATE_MULTIPLIER*waveform.nAvgBytesPerSec);
                    MessageBox (hwnd, szOutOfMemory, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
                    return TRUE ;
               }

               if (waveInOpen (&hWaveIn, WAVE_MAPPER, &waveform, 
                               (DWORD) hwnd, 0, CALLBACK_WINDOW))
               {
                    sprintf(Line,szOpenErrorDev,wic.dwFormats);
                    free (pBuffer1) ;
                    free (pBuffer2) ;
                    MessageBeep (MB_ICONEXCLAMATION) ;
                    MessageBox (hwnd, Line, szAppName,
                                      MB_ICONEXCLAMATION | MB_OK) ;
					return FALSE;
               }

                    // Set up headers and prepare them
               
               pWaveHdr1->lpData          = (char*)pBuffer1 ;
               pWaveHdr1->dwBufferLength  = (RATE_MULTIPLIER*waveform.nAvgBytesPerSec);
               pWaveHdr1->dwBytesRecorded = 0 ;
               pWaveHdr1->dwUser          = 0 ;
               pWaveHdr1->dwFlags         = 0 ;
               pWaveHdr1->dwLoops         = 1 ;
               pWaveHdr1->lpNext          = NULL ;
               pWaveHdr1->reserved        = 0 ;

               waveInPrepareHeader (hWaveIn, pWaveHdr1, sizeof (WAVEHDR)) ;
          
               pWaveHdr2->lpData          = (char*)pBuffer2 ;
               pWaveHdr2->dwBufferLength  = (RATE_MULTIPLIER*waveform.nAvgBytesPerSec);
               pWaveHdr2->dwBytesRecorded = 0 ;
               pWaveHdr2->dwUser          = 0 ;
               pWaveHdr2->dwFlags         = 0 ;
               pWaveHdr2->dwLoops         = 1 ;
               pWaveHdr2->lpNext          = NULL ;
               pWaveHdr2->reserved        = 0 ;
          
               waveInPrepareHeader (hWaveIn, pWaveHdr2, sizeof (WAVEHDR)) ;
               return TRUE ;
          case IDC_PLAY:
               if ((bRecording) || (bPlaying) || (!dwDataLength))
               {
                   return TRUE;
               }
               
               // Open waveform audio for output
               // NOTE: WAVEFORMATEX waveform should have already
               //       been initialized to the proper values upon
               //       file load or record-session
               if (waveOutOpen (&hWaveOut, WAVE_MAPPER, &waveform, 
                                (DWORD) hwnd, 0, CALLBACK_WINDOW))
               {
                    MessageBeep (MB_ICONEXCLAMATION) ;
                    MessageBox (hwnd, szOpenError, szAppName,
                         MB_ICONEXCLAMATION | MB_OK) ;
               }
               // force window to repaint and null out data
               InitPoints();
               RepaintWindow();
               return TRUE ;
          case IDC_PAUSE:
                    // Pause or restart output
               if (bPlaying)
               {
                  if (!bPaused)
                  {
                       waveOutPause (hWaveOut) ;
                       SetDlgItemText (hwnd, IDC_PAUSE, TEXT ("Resume")) ;
                       bPaused = TRUE ;
                       Announce("Paused",dwCurrentPos,dwDataLength);
                  }
                  else
                  {
                       waveOutRestart (hWaveOut) ;
                       SetDlgItemText (hwnd, IDC_PAUSE, TEXT ("Pause")) ;
                       bPaused = FALSE ;
                       Announce("Playing",dwCurrentPos,dwDataLength);
                  }
               }
               return TRUE ;
          case IDC_STOP:
               if (bRecording)
               {
                  // Reset input to return last buffer
                  waveInReset (hWaveIn) ;
                  bEnding = TRUE ;
               }
               else 
                  if (bPlaying)
                  {
                  
                     // Reset output for close preparation
                     waveOutReset (hWaveOut) ;
                     bEnding = TRUE ;
                  } 
                  else
                  {
                     dwCurrentPos = 0;
                     Announce("Ready",dwCurrentPos,dwDataLength);

                  }
               return TRUE ;               
           }
      case WM_TIMER:
           {
           int i;
              // Loop for 2-note polyphony
               for (i = 0 ; i < 2 ; i++)
               {
                    // Note Off messages for previous note
                  if (iIndex != 0 && noteseq[iIndex - 1].iNote[i] != -1)
                  {
                       MidiOutMessage (hMidiOut, 0x80,  0,
                                       noteseq[iIndex - 1].iNote[i], 0) ;
                  }
                    // Note On messages for new note               
                  if (iIndex != sizeof (noteseq) / sizeof (noteseq[0]) &&
                       noteseq[iIndex].iNote[i] != -1)
                  {
                    MidiOutMessage (hMidiOut, 0x90,  0,
                                    noteseq[iIndex].iNote[i], 127) ;
                  }
               }
               if (iIndex != sizeof (noteseq) / sizeof (noteseq[0]))
               {
                  SetTimer (hwnd, ID_TIMER, noteseq[iIndex++].iDur - 1, NULL) ;
               }
               else
               {
                  midiOutReset (hMidiOut) ;
                  midiOutClose (hMidiOut) ;
                  KillTimer (hwnd, ID_TIMER) ;
               }
               return 0;
          }
             
     case MM_WIM_OPEN:
               // Shrink down the save buffer
          pSaveBuffer = realloc (pSaveBuffer, 1) ;
          
               // Enable and disable Buttons
          SetDialogState(hwnd,hwndMessage,hwndPlayback,FALSE,FALSE,TRUE,FALSE,IDC_STOP,dwCurrentPos,dwDataLength);
          
               // Add the buffers
          waveInAddBuffer (hWaveIn, pWaveHdr1, sizeof (WAVEHDR)) ;
          waveInAddBuffer (hWaveIn, pWaveHdr2, sizeof (WAVEHDR)) ;
          
               // Begin sampling
          bRecording   = TRUE ;
          bEnding      = FALSE ;
          bNeedSave    = FALSE ;
          dwCurrentPos = 0;
          dwDataLength = 0 ;
          waveInStart (hWaveIn) ;
          Announce("Recording",dwDataLength,0);

          return TRUE ;
          
     case MM_WIM_DATA:
         
               // Reallocate save buffer memory
          
          pNewBuffer = realloc (pSaveBuffer, dwDataLength +
                                   ((PWAVEHDR) lParam)->dwBytesRecorded) ;
          
          if (pNewBuffer == NULL)
          {
               waveInClose (hWaveIn) ;
               MessageBeep (MB_ICONEXCLAMATION) ;
               sprintf(Line,szOutOfMemory,dwDataLength +
                                          ((PWAVEHDR) lParam)->dwBytesRecorded);
               MessageBox (hwnd, Line, szAppName,
                                 MB_ICONEXCLAMATION | MB_OK) ;
               return TRUE ;
          }
          
          pSaveBuffer = pNewBuffer ;
          CopyMemory (pSaveBuffer + dwDataLength, ((PWAVEHDR) lParam)->lpData,
                         ((PWAVEHDR) lParam)->dwBytesRecorded) ;
          
          dwDataLength += ((PWAVEHDR) lParam)->dwBytesRecorded ;
          Announce("Recording",dwDataLength,0);

          if (bEnding)
          {
               waveInClose (hWaveIn) ;
               return TRUE ;
          }
          
               // Send out a new buffer
          
          waveInAddBuffer (hWaveIn, (PWAVEHDR) lParam, sizeof (WAVEHDR)) ;
          return TRUE ;
          
     case MM_WIM_CLOSE:
               // Free the buffer memory

          waveInUnprepareHeader (hWaveIn, pWaveHdr1, sizeof (WAVEHDR)) ;
          waveInUnprepareHeader (hWaveIn, pWaveHdr2, sizeof (WAVEHDR)) ;

          free (pBuffer1) ;
          free (pBuffer2) ;
          
               // Enable and disable buttons          
          if (dwDataLength > 0)
          { 
               bNeedSave = TRUE;
               SetDialogState(hwnd,hwndMessage,hwndPlayback,TRUE,FALSE,TRUE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength);
          }
          else
          {
               bNeedSave = FALSE;
               SetDialogState(hwnd,hwndMessage,hwndPlayback,FALSE,FALSE,TRUE,TRUE,IDC_RECORD,dwCurrentPos,dwDataLength);
          }
          bRecording = FALSE ;
         
          if (bTerminating)
               SendMessage (hwnd, WM_SYSCOMMAND, SC_CLOSE, 0L) ;
          else
          {
              // Do proper analysis of recorded data
              // analyze file read
              SetWindowText(hwndMessage,TEXT("Analyzing File"));
              Analyze(pSaveBuffer,dwDataLength);
                     
              // set button states, but preserve loaded content
              SetDialogState(hwnd,hwndMessage,hwndPlayback,TRUE,FALSE,TRUE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength);
          }
          
          return TRUE ;
          
     case MM_WOM_OPEN:
               // Enable and disable buttons
          SetDialogState(hwnd,hwndMessage,hwndPlayback,FALSE,TRUE,TRUE,FALSE,IDC_STOP,dwCurrentPos,dwDataLength);
          
          if (dwCurrentPos < dwDataLength)
          {
             // initiate double buffered output 
             QueueWaveOut(hWaveOut,pWaveHdr1,pSaveBuffer,&dwCurrentPos,dwDataLength);
             QueueWaveOut(hWaveOut,pWaveHdr2,pSaveBuffer,&dwCurrentPos,dwDataLength);                    
             bEnding = FALSE ;
             bPlaying = TRUE ;
             Announce("Playing",dwCurrentPos,dwDataLength);
          }
          else
          {
             waveOutClose (hWaveOut) ;
          }
          return TRUE ;

         
     case MM_WOM_DONE:

          // Send next block until done
          if (!bEnding && (dwCurrentPos < dwDataLength)) 
          {

              // restart double buffer with current buffer
              QueueWaveOut(hWaveOut,(PWAVEHDR) lParam,pSaveBuffer,&dwCurrentPos,dwDataLength);
              Announce("Playing",dwCurrentPos,dwDataLength);
          }
          else
          {
             // deallocate double buffer
             waveOutUnprepareHeader (hWaveOut, (PWAVEHDR) lParam, sizeof (WAVEHDR)) ;
             ((PWAVEHDR) lParam)->dwBufferLength = 0;
             
             // when second buffer arrives, close device
             if ((pWaveHdr1->dwBufferLength == 0) && (pWaveHdr2->dwBufferLength == 0))
             {
                waveOutClose (hWaveOut) ;
             }
          }
          return TRUE ;
     case MM_WOM_CLOSE:
          // if played to the finish, set to start over
          if (dwCurrentPos >= dwDataLength)
          {
             dwCurrentPos = 0;
          }
          
          // Enable and disable buttons
          if (dwDataLength > 0)
          {
             SetDialogState(hwnd,hwndMessage,hwndPlayback,TRUE,FALSE,TRUE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength); 
          }
          else
          {
             SetDialogState(hwnd,hwndMessage,hwndPlayback,FALSE,FALSE,FALSE,TRUE,IDC_PLAY,dwCurrentPos,dwDataLength); 
          }         
          SetDlgItemText (hwnd, IDC_PAUSE, TEXT ("Pause")) ;
          bPaused = FALSE ;
          bPlaying = FALSE ;

          if (bTerminating)
               SendMessage (hwnd, WM_SYSCOMMAND, SC_CLOSE, 0L) ;
          
          return TRUE ;
          
     case WM_SYSCOMMAND:
          switch (LOWORD (wParam))
          {
          case SC_CLOSE:
               if (bRecording)
               {
                    bTerminating = TRUE ;
                    bEnding = TRUE ;
                    waveInReset (hWaveIn) ;
                    return TRUE ;
               }
               
               if (bPlaying)
               {
                    bTerminating = TRUE ;
                    bEnding = TRUE ;
                    waveOutReset (hWaveOut) ;
                    return TRUE ;
               }
               
               free (pWaveHdr1) ;
               free (pWaveHdr2) ;
               free (pSaveBuffer) ;
               EndDialog (hwnd, 0) ;
               return TRUE ;
          }
          break ;
          // Vertical scroll is the SPINCONTROL WIDGET
     case WM_VSCROLL : 
          Value = GetDlgItemInt(hwnd,IDC_BUDDY,&Error,FALSE);
          switch (LOWORD (wParam))
          {
          case SB_LINEDOWN : if (Value>0)  
                             {
                                Value = Value-1; 
                                SetDlgItemInt(hwnd,IDC_BUDDY,Value,FALSE);
                                Frequency_Threshold = (double) Value/100.0;
                                SendMessage (hwnd, WM_PAINT, 0, 0) ;
                             }
                             break;
          case SB_LINEUP   : if (Value<50) 
                             {
                                Value = Value+1;
                                SetDlgItemInt(hwnd,IDC_BUDDY,Value,FALSE);
                                Frequency_Threshold = (double) Value/100.0;
                                SendMessage (hwnd, WM_PAINT, 0, 0) ;
                             }
                             break;
          default: break;
          }
          break;
          
          // Horizontal scroll is the playback widget
     case WM_HSCROLL :          
          switch (LOWORD (wParam))
          {
          case SB_PAGEDOWN :
               dwCurrentPos += waveform.nAvgBytesPerSec*10;
               if (dwCurrentPos > dwDataLength)
               {
                  dwCurrentPos = dwDataLength;
               }
               break;
          case SB_LINEDOWN :
               dwCurrentPos += waveform.nAvgBytesPerSec;
               if (dwCurrentPos > dwDataLength)
               {
                  dwCurrentPos = dwDataLength;
               }
               break;
               
          case SB_PAGEUP :
               dwCurrentPos -= waveform.nAvgBytesPerSec*10;
               if (dwCurrentPos < 0)
               {
                  dwCurrentPos = 0;
               }
               break;
          case SB_LINEUP :
               dwCurrentPos -= waveform.nAvgBytesPerSec;
               if (dwCurrentPos < 0) 
               {
                  dwCurrentPos = 0;
               }
               break;
          case SB_TOP :
               dwCurrentPos = 0;
               break;               
          case SB_BOTTOM :
               dwCurrentPos = dwDataLength;
               break;               
          case SB_THUMBPOSITION :
          case SB_THUMBTRACK :
               // decode scroll position to buffer position
               // scroll only to integer second boundaries
               // this will keep from replaying from boundaries
               // different than those analyzed
               // which might yield different analysis results for
               // short interval spikes if the fall on an analysis boundary
               dwCurrentPos = (int)(
                                     (double) HIWORD (wParam) *
                                     (double)((int) dwDataLength /
                                              (int) waveform.nAvgBytesPerSec
                                             ) *
                                     (double) waveform.nAvgBytesPerSec /
                                     (double) PLAYBACK_RESOLUTION ) ;
               break ;
          default :
               return FALSE ;
          }   
        Announce("Play Position",dwCurrentPos,dwDataLength);
        return TRUE;
     }
     return FALSE ;
}
